import click
import functools
import pathlib
import re
import torch
import traceback
import yaml
from sd_mecha import extensions
from sd_mecha.recipe_merger import RecipeMerger
from sd_mecha.recipe_nodes import ParameterResolverVisitor
from sd_mecha.recipe_serializer import deserialize, deserialize_path, serialize
from sd_mecha.user_error import UserError
from typing import Optional, Tuple, List


DTYPE_MAPPING = {
    "bf16": torch.bfloat16,
    "fp16": torch.float16,
    "fp32": torch.float32,
    "fp64": torch.float64,
}


def except_fallback(f):
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        debug = kwargs['debug']
        if debug:
            click.echo("Debug mode active.", err=True)
        try:
            return f(*args, **kwargs)
        except UserError as e:
            print_exception(debug, e)
        except Exception as e:
            print_exception(debug, e)
            if not debug:
                click.echo("An unexpected error occurred. Run with --debug to see more details.", err=True)
    return wrapped


def print_exception(debug: bool, e: Exception):
    if debug:
        click.echo(traceback.format_exc(), err=True)
    else:
        click.echo(str(e), err=True)


@click.group()
def main():
    pass


@click.command(help="Merge models following a recipe")
@click.argument("recipe", type=pathlib.Path)
@click.option("--models-directory", type=pathlib.Path, envvar="MECHA_MODELS_DIR", default=pathlib.Path(), help="Base directory for .safetensors checkpoints.")
@click.option("--output", "-o", type=pathlib.Path, default=None, help="Output file for the merge result.")
@click.option("--threads", "-j", default=1, help="Number of keys to merge in parallel.")
@click.option("--device", "-d", default="cpu", help="Device used for merging.")
@click.option("--dtype", "-p", type=click.Choice(list(DTYPE_MAPPING.keys())), default="fp32", help="Work precision.")
@click.option("--save-dtype", "-t", type=click.Choice(list(DTYPE_MAPPING.keys())), default="fp16", help="Save precision.")
@click.option("--debug", is_flag=True, help="Print the stacktrace when an error occurs.")
@except_fallback
def merge(
    recipe: pathlib.Path,
    models_directory: pathlib.Path,
    output: pathlib.Path,
    threads: int,
    device: str,
    dtype: str,
    save_dtype: str,
    debug: bool,
):
    if output is None:
        output = models_directory / "merge.safetensors"

    recipe = deserialize(recipe)
    scheduler = RecipeMerger(
        models_dir=models_directory,
        default_device=device,
        default_dtype=DTYPE_MAPPING[dtype],
    )
    scheduler.merge_and_save(
        recipe,
        output=output,
        save_dtype=DTYPE_MAPPING[save_dtype],
        threads=threads
    )


@click.command(help="Compose recipes together to form a larger recipe")
@click.argument("base_recipe", type=pathlib.Path)
@click.option("--models-directory", type=pathlib.Path, envvar="MECHA_MODELS_DIR", default=pathlib.Path(), help="Base directory for .safetensors checkpoints.")
@click.option("arguments", "--argument", "-p", type=(str, str), multiple=True, help="Recipe or model argument to pass to the base recipe.")
@click.option("--output", "-o", type=pathlib.Path, default=None, help="Output file of the composed recipe.")
@click.option("--debug", is_flag=True, help="Print the stacktrace when an error occurs.")
@except_fallback
def compose(
    base_recipe: pathlib.Path,
    models_directory: pathlib.Path,
    arguments: List[Tuple[str, str]],
    output: Optional[pathlib.Path],
    debug: bool,
):
    if output is None:
        output = pathlib.Path("merge.mecha")

    base_recipe = deserialize(base_recipe)
    arguments = {
        k: deserialize_path(v, models_directory)
        for k, v in arguments
    }
    composed_recipe = base_recipe.accept(ParameterResolverVisitor(arguments))
    composed_recipe = serialize(composed_recipe)
    with open(output, "w") as f:
        f.write(composed_recipe)


@click.command(help="Show the available blocks and classes of model architectures")
@click.argument("model_arch", type=click.Choice(extensions.model_arch.get_all() + [""], case_sensitive=False), default="")
@click.option("--verbose", "-v", is_flag=True)
@click.option("--debug", is_flag=True, help="Print the stacktrace when an error occurs.")
@except_fallback
def info(
    model_arch: str,
    verbose: bool,
    debug: bool,
):
    if not model_arch:
        print(yaml.dump({
            "Available architectures": [
                arch
                if not verbose else
                f"{arch} (registered at {extensions.model_arch.resolve(arch).location})"
                for arch in extensions.model_arch.get_all()
            ]
        }))
        return

    model_arch = extensions.model_arch.resolve(model_arch)
    print(yaml.dump({
        f'Component "{component}"': {
            "Blocks":
                sorted(list({
                    b.split("_", 3)[3]
                    for bs in model_arch.blocks.values()
                    for b in bs
                    if f"_{component}_block_" in b
                }), key=natural_sort_key)
                if not verbose else
                dict(sorted([
                    (b.split("_", 3)[3], sorted([
                        k for k in model_arch.blocks
                        if b in model_arch.blocks[k]
                    ], key=natural_sort_key))
                    for bs in model_arch.blocks.values()
                    for b in bs
                    if f"_{component}_block_" in b
                ], key=lambda t: natural_sort_key(t[0]))),
            "Classes":
                sorted(list({
                    c.split("_", 3)[3]
                    for cs in model_arch.classes.values()
                    for c in cs
                    if f"_{component}_class_" in c
                }), key=natural_sort_key)
                if not verbose else
                dict(sorted([
                    (c.split("_", 3)[3], sorted([
                        k for k in model_arch.classes
                        if c in model_arch.classes[k]
                    ], key=natural_sort_key))
                    for cs in model_arch.classes.values()
                    for c in cs
                    if f"_{component}_class_" in c
                ], key=lambda t: natural_sort_key(t[0]))),
        }
        for component in model_arch.components
    }, sort_keys=False))


def natural_sort_key(s):
    return [
        int(text)
        if text.isdigit()
        else text.lower()
        for text in re.split('([0-9]+)', s)
    ]


main.add_command(merge)
main.add_command(compose)
main.add_command(info)


if __name__ == "__main__":
    main()



import dataclasses
import functools
import pathlib

import fuzzywuzzy.process
import traceback
import torch
from sd_mecha import extensions
from sd_mecha.merge_space import MergeSpace
from sd_mecha.extensions.model_arch import ModelArch
from sd_mecha.streaming import DTYPE_REVERSE_MAPPING, DTYPE_MAPPING, InSafetensorsDict
from torch._subclasses.fake_tensor import FakeTensorMode
from typing import Mapping, Optional, Iterable, Protocol, Tuple


class ModelTypeCallback(Protocol):
    # Define types here, as if __call__ were a function (ignore self).
    def __call__(self, state_dict: Mapping[str, torch.Tensor], key: str, **kwargs) -> torch.Tensor:
        ...


@dataclasses.dataclass
class ModelType:
    __f: ModelTypeCallback
    identifier: str
    merge_space: MergeSpace
    needs_header_conversion: bool
    strict_suffixes: bool
    key_suffixes: Optional[Tuple[str]]
    location: str

    def get_tensor(self, sd_path: pathlib.Path | str, state_dict: Mapping[str, torch.Tensor], key: str) -> torch.Tensor:
        return self.__f(state_dict, key, sd_path=sd_path)

    def convert_header(self, sd_path: pathlib.Path | str, state_dict: InSafetensorsDict | Mapping[str, torch.Tensor], model_arch: ModelArch):
        def _create_header(fake_state_dict):
            data_offsets = 0
            return {
                k: {
                    "dtype": DTYPE_REVERSE_MAPPING[t.dtype][0],
                    "shape": list(t.shape),
                    "data_offsets": [data_offsets, (data_offsets := data_offsets + t.numel() * t.element_size())],
                }
                for k, t in fake_state_dict.items()
            }

        if isinstance(state_dict, InSafetensorsDict):
            if not self.needs_header_conversion:
                return {
                    k: v
                    for k, v in state_dict.header.items()
                    if k != "__metadata__"
                }

            def _create_fake_tensor(*shape, dtype):
                return torch.empty(shape, dtype=dtype)

            if self.strict_suffixes:
                for sd_key in state_dict:
                    if sd_key == "__metadata__":
                        continue

                    if not sd_key.endswith(self.key_suffixes):
                        raise RuntimeError(
                            f"cannot load a non-{self.identifier} network using the {self.identifier} model type. ({sd_path})\n" +
                            f"found key not matching allowed suffixes {self.key_suffixes}: {sd_key}"
                        )

            with FakeTensorMode():
                fake_state_dict = {
                    k: _create_fake_tensor(*h["shape"], dtype=DTYPE_MAPPING[h["dtype"]][0])
                    for k, h in state_dict.header.items()
                    if k != "__metadata__" and (self.key_suffixes is None or k.endswith(self.key_suffixes))
                }
                converted_state_dict = {}
                for k in model_arch.keys:
                    try:
                        converted_state_dict[k] = self.get_tensor(sd_path, fake_state_dict, k)
                    except KeyError:
                        continue

                return _create_header(converted_state_dict)
        else:
            return _create_header(state_dict)


def register_model_type(
    *,
    merge_space: MergeSpace,
    identifier: Optional[str] = None,
    model_archs: str | Iterable[str] = ("__default__",),
    needs_header_conversion: bool = True,
    strict_suffixes: bool = False,
    key_suffixes: Optional[Iterable[str]] = None,
):
    stack_frame = traceback.extract_stack(None, 2)[0]
    partial = functools.partial(
        __register_model_type_impl,
        identifier=identifier,
        merge_space=merge_space,
        model_archs=model_archs,
        needs_header_conversion=needs_header_conversion,
        strict_suffixes=strict_suffixes,
        key_suffixes=key_suffixes,
        stack_frame=stack_frame,
    )
    return partial


def __register_model_type_impl(
    f: ModelTypeCallback,
    *,
    identifier: Optional[str],
    merge_space: MergeSpace,
    model_archs: str | Iterable[str],
    needs_header_conversion: bool,
    strict_suffixes: bool,
    key_suffixes: Optional[Iterable[str]],
    stack_frame: traceback.FrameSummary,
):
    if identifier is None:
        identifier = f.__name__

    if isinstance(model_archs, str):
        model_archs = [model_archs]

    if not model_archs:
        raise ValueError(f"cannot register model type '{identifier}' without an architecture")

    if key_suffixes is not None:
        key_suffixes = tuple(key_suffixes)

    model_archs = [
        extensions.model_arch.resolve(model_arch).identifier
        if model_arch != "__default__"
        else model_arch
        for model_arch in model_archs
    ]
    if identifier in _model_types_registry and (not model_archs or any(model_arch in _model_types_registry[identifier] for model_arch in model_archs)):
        existing_location = _model_types_registry[identifier].location
        raise ValueError(f"model type extension '{identifier}' is already registered at {existing_location}.")

    location = f"{stack_frame.filename}:{stack_frame.lineno}"
    if identifier not in _model_types_registry:
        _model_types_registry[identifier] = {}
    for model_arch in model_archs:
        _model_types_registry[identifier][model_arch] = ModelType(f, identifier, merge_space, needs_header_conversion, strict_suffixes, key_suffixes, location)

    return f


_model_types_registry = {}


def resolve(identifier: str, arch_identifier: str) -> ModelType:
    try:
        related_model_types = _model_types_registry[identifier]
        res = related_model_types.get(arch_identifier)
        if res is None:
            res = related_model_types.get("__default__")
        if res is None:
            raise KeyError(identifier)
        return res
    except KeyError as e:
        suggestion = fuzzywuzzy.process.extractOne(str(e), _model_types_registry.keys())[0]
        raise ValueError(f"unknown model type: {e}. Nearest match is '{suggestion}'")



import re
import fuzzywuzzy.process
from typing import Dict, Optional, List
from sd_mecha import extensions
from sd_mecha.extensions.model_arch import ModelArch


Hyper = int | float | Dict[str, int | float]


def get_hyper(hyper: Hyper, key: str, model_arch: ModelArch) -> int | float:
    if isinstance(hyper, (float, int)):
        return hyper
    elif isinstance(hyper, dict):
        hyper_ids = model_arch.classes[key] | model_arch.blocks[key]
        result = 0.0
        total = 0
        for hyper_id in hyper_ids:
            value = hyper.get(hyper_id)
            if value is not None:
                result += value
                total += 1

        if total > 0:
            return result / total

        for hyper_id in hyper_ids:
            for component in model_arch.components:
                if f"_{component}_" not in hyper_id:
                    continue

                try:
                    return hyper[model_arch.identifier + "_" + component + "_default"]
                except KeyError:
                    continue

        return 0
    else:
        raise TypeError(f"Hyperparameter must be a float or a dictionary, not {type(hyper).__name__}")


def validate_hyper(hyper: Hyper, model_arch: Optional[ModelArch]) -> Hyper:
    if isinstance(hyper, dict):
        if model_arch is None:
            raise ValueError("Abstract recipes (with recipe parameters) cannot specify component-wise hyperparameters")

        user_keys = model_arch.user_keys()
        for key in hyper.keys():
            if key not in user_keys and not key.endswith("_default"):
                suggestion = fuzzywuzzy.process.extractOne(key, user_keys)[0]
                raise ValueError(f"Unsupported dictionary key '{key}'. Nearest match is '{suggestion}'")
    elif isinstance(hyper, (int, float)):
        return hyper
    else:
        raise TypeError(f"Hyperparameter must be a float or a dictionary, not {type(hyper).__name__}")


def blocks(model_arch: str | ModelArch, model_component: str, *args, strict: bool = True, **kwargs) -> Hyper:
    """
    Generate block hyperparameters for a model version.
    Either positional arguments or keyword arguments can be used, but not both at the same time.
    If positional blocks are used, they must all be specified.
    The CLI has a command to help determine the names of the blocks without guessing:

    ```
    python -m sd_mecha info <model_arch>
    ```

    where `<model_arch>` is the version identifier of the model used. (i.e. "sd1", "sdxl", etc.)

    :param model_arch: the architecture of the model for which to generate block hyperparameters
    :param model_component: the model component for which the block-wise hyperparameters are intended for (i.e. "unet", "txt", "txt2", etc.)
    :param strict: if blocks are passed through *args, determines whether the number of blocks must match exactly the maximum number of blocks for the selected model component
    :param args: positional block hyperparameters
    :param kwargs: keyword block hyperparameters
    :return: block-wise hyperparameters
    """
    if isinstance(model_arch, str):
        model_arch = extensions.model_arch.resolve(model_arch)

    if args and kwargs:
        raise ValueError("`args` and `kwargs` cannot be used at the same time. Either use positional or keyword arguments, but not both.")
    if args:
        identifiers = list(
            k for k in model_arch.user_keys()
            if "_" + model_component + "_block_" in k
        )
        if strict and len(args) != len(identifiers):
            raise ValueError(f"blocks() got {len(args)} block{'s' if len(args) > 1 else ''} but {len(identifiers)} are expected. Use keyword arguments to pass only a few (i.e. 'in0=1, out3=0.5, ...') or pass strict=False.")

        identifiers.sort(key=natural_sort_key)
        return dict(zip(identifiers, args))
    return {
        model_arch.identifier + "_" + model_component + "_block_" + k: v
        for k, v in kwargs.items()
    }


def natural_sort_key(s):
    return [int(text) if text.isdigit() else text.lower() for text in re.split('([0-9]+)', s)]


def classes(model_arch: str | ModelArch, model_component: str, **kwargs) -> Hyper:
    """
    Generate class hyperparameters for a model version.
    The CLI has a command to help determine the names of the classes for a given model version without guessing:

    ```
    python -m sd_mecha info <model_arch>
    ```

    where `<model_arch>` is the version identifier of the model used. (i.e. "sd1", "sdxl", etc.)

    :param model_arch: the architecture of the model for which to generate class hyperparameters
    :param model_component: the model component for which the class-wise hyperparameters are intended for (i.e. "unet", "txt", "txt2", etc.)
    :param kwargs: class hyperparameters by name
    :return: class-wise hyperparameters
    """
    if isinstance(model_arch, str):
        model_arch = extensions.model_arch.resolve(model_arch)

    return {
        model_arch.identifier + "_" + model_component + "_class_" + k: v
        for k, v in kwargs.items()
    }


def default(model_arch: str | ModelArch, model_components: Optional[str | List[str]] = None, value: int | float = 0) -> Hyper:
    if isinstance(model_arch, str):
        model_arch = extensions.model_arch.resolve(model_arch)

    if not model_components:
        model_components = model_arch.components
    elif isinstance(model_components, str):
        model_components = [model_components]

    return {
        model_arch.identifier + "_" + model_component + "_default": value
        for model_component in model_components
    }



import contextlib
import dataclasses
import functools
import json
import operator
import pathlib
import struct
import threading
import torch
import warnings
from typing import Optional, Set
from tqdm import tqdm


class InSafetensorsDict:
    def __init__(self, file_path: pathlib.Path, buffer_size):
        if not file_path.suffix == ".safetensors":
            raise ValueError(f"Model type not supported: {file_path} (only safetensors are supported)")

        self.default_buffer_size = buffer_size
        self.file_path = file_path
        self.file = open(file_path, mode='rb', buffering=0)
        self.header_size, self.header = self._read_header()
        self.buffer = bytearray()
        self.buffer_start_offset = 8 + self.header_size
        self.lock = threading.Lock()

    def __del__(self):
        self.close()

    def __getitem__(self, key):
        if key not in self.header or key == "__metadata__":
            raise KeyError(key)
        return self._load_tensor(key)

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self.header)

    def close(self):
        self.file.close()
        self.buffer = None
        self.header = None

    def keys(self):
        return (
            key
            for key in self.header.keys()
            if key != "__metadata__"
        )

    def values(self):
        for key in self.keys():
            yield self[key]

    def items(self):
        for key in self.keys():
            yield key, self[key]

    def _read_header(self):
        header_size_bytes = self.file.read(8)
        header_size = struct.unpack('<Q', header_size_bytes)[0]
        header_json = self.file.read(header_size).decode('utf-8').strip()
        header = json.loads(header_json)

        # sort by memory order to reduce seek time
        sorted_header = dict(sorted(header.items(), key=lambda item: item[1].get('data_offsets', [0])[0]))
        return header_size, sorted_header

    def _ensure_buffer(self, start_pos, length):
        if start_pos < self.buffer_start_offset or start_pos + length > self.buffer_start_offset + len(self.buffer):
            self.file.seek(start_pos)
            necessary_buffer_size = max(self.default_buffer_size, length)
            if len(self.buffer) < necessary_buffer_size:
                self.buffer = bytearray(necessary_buffer_size)
            else:
                self.buffer = self.buffer[:necessary_buffer_size]

            self.file.readinto(self.buffer)
            self.buffer_start_offset = start_pos

    def _load_tensor(self, tensor_name):
        tensor_info = self.header[tensor_name]
        offsets = tensor_info['data_offsets']
        dtype, dtype_bytes = DTYPE_MAPPING[tensor_info['dtype']]
        shape = tensor_info['shape']
        total_bytes = offsets[1] - offsets[0]
        absolute_start_pos = 8 + self.header_size + offsets[0]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            with self.lock:
                self._ensure_buffer(absolute_start_pos, total_bytes)
                buffer_offset = absolute_start_pos - self.buffer_start_offset
                return torch.frombuffer(self.buffer, count=total_bytes // dtype_bytes, offset=buffer_offset, dtype=dtype).reshape(shape)


@dataclasses.dataclass
class OutSafetensorsDictThreadState:
    buffer: bytearray
    memory_used: int = dataclasses.field(default=0)
    sub_header: dict = dataclasses.field(default_factory=dict)


class OutSafetensorsDict:
    def __init__(
        self,
        file_path: pathlib.Path,
        template_header: dict,
        keys_to_merge: Set[str],
        mecha_recipe: str,
        minimum_buffer_size: int,
        save_dtype: torch.dtype,
    ):
        self.thread_states = {}
        self.lock = threading.Lock()

        self.header = {
            "__metadata__": {"mecha_recipe": mecha_recipe}
        }
        self.file = file_path.open("wb", buffering=0)
        self.file_path = file_path
        self.flushed_size = 0
        self.minimum_buffer_size = minimum_buffer_size

        self.max_header_size = self._init_buffer(template_header, keys_to_merge, save_dtype)

    def __del__(self):
        self.file.close()
        self.thread_states = None
        self.header = None

    def __setitem__(self, key: str, tensor: torch.Tensor):
        tid = threading.current_thread().ident
        if tid not in self.thread_states:
            self.thread_states[tid] = OutSafetensorsDictThreadState(bytearray(self.minimum_buffer_size))

        state = self.thread_states[tid]

        tensor_bytes = tensor.cpu().numpy().tobytes()
        tensor_size = len(tensor_bytes)

        if tensor_size > len(state.buffer) - state.memory_used:
            self._flush_buffer(state, next_tensor_size=tensor_size)

        local_offset = state.memory_used
        state.buffer[state.memory_used:state.memory_used + tensor_size] = tensor_bytes
        state.memory_used += tensor_size

        state.sub_header[key] = {
            "dtype": DTYPE_REVERSE_MAPPING[tensor.dtype][0],
            "shape": list(tensor.shape),
            "data_offsets": [local_offset, local_offset + tensor_size]
        }

    def __len__(self):
        return len(self.header)

    def _init_buffer(self, template_header: dict, keys_to_merge: Set[str], save_dtype: torch.dtype) -> int:
        def get_dtype(k):
            return DTYPE_REVERSE_MAPPING[save_dtype][0] if k in keys_to_merge else template_header[k]["dtype"]

        def get_dtype_size(k):
            return DTYPE_REVERSE_MAPPING[save_dtype][1] if k in keys_to_merge else DTYPE_MAPPING[template_header[k]["dtype"]][1]

        def get_width(k):
            return functools.reduce(operator.mul, template_header[k]["shape"], 1) * get_dtype_size(k)

        keys = sorted(
            list(k for k in template_header.keys() if k != "__metadata__"),
            key=get_width,
            reverse=True,  # simulate worst case: maximize space taken by order
        )
        data_offset = 0
        dummy_header = {
            k: {
                "dtype": get_dtype(k),
                "shape": template_header[k]["shape"],
                "data_offsets": [data_offset, (data_offset := data_offset + get_width(k))],
            }
            for k in keys
        }

        dummy_header["__metadata__"] = self.header["__metadata__"]
        header_json = json.dumps(dummy_header, separators=(',', ':')).encode('utf-8')
        max_header_size = len(header_json)
        self.file.seek(8 + max_header_size)  # Reserve space for the header
        return max_header_size

    def _flush_buffer(self, state: OutSafetensorsDictThreadState, next_tensor_size: Optional[int] = None, close: bool = False):
        if not close:
            lock_obj = self.lock
        else:
            lock_obj = contextlib.nullcontext()

        with lock_obj:
            self.file.write(state.buffer[:state.memory_used])
            buffer_offset = self.flushed_size
            self.flushed_size += state.memory_used
            state.memory_used = 0

        if next_tensor_size is not None:
            required_buffer_size = max(self.minimum_buffer_size, next_tensor_size)
            if len(state.buffer) < required_buffer_size:
                state.buffer = bytearray(required_buffer_size)
            else:
                state.buffer = state.buffer[:required_buffer_size]

        global_sub_header = {
            k: {
                attr: val
                if attr != "data_offsets"
                else (val[0] + buffer_offset, val[1] + buffer_offset)
                for attr, val in v.items()
            }
            for k, v in state.sub_header.items()
        }
        self.header.update(global_sub_header)
        state.sub_header.clear()
        if close:
            state.buffer = b""

    def close(self):
        with self.lock:
            for state in self.thread_states.values():
                self._flush_buffer(state, close=True)

            header_json = json.dumps(self.header, separators=(',', ':')).encode('utf-8')
            header_size = len(header_json)
            overhead = self.max_header_size - header_size

            if overhead < 0:
                # not enough space. we have to move the entire data section by `-overhead`
                # this should never happen, but it's here just in case as a fallback
                data_offset = -overhead
                old_data_section = 8 + self.max_header_size
                old_file_end = 8 + self.max_header_size + self.flushed_size
                new_file_end = 8 + header_size + self.flushed_size
                self.file.truncate(new_file_end)

                # close and reopen the file in read-write mode
                self.file.close()
                self.file = open(self.file_path, "rb+")

                # move data in chunks from the end to avoid overwriting
                for chunk_end in tqdm(range(old_file_end, old_data_section, -self.minimum_buffer_size), desc="Reallocating data section"):
                    chunk_start = max(chunk_end - self.minimum_buffer_size, old_data_section)
                    chunk_size = chunk_end - chunk_start
                    self.file.seek(chunk_start)
                    data = self.file.read(chunk_size)

                    # calculate the new position and write the chunk
                    self.file.seek(chunk_start + data_offset)
                    self.file.write(data)

                # we made just enough space for the header
                overhead = 0

            self.file.seek(0)
            self.file.write(struct.pack('<Q', max(self.max_header_size, header_size)))
            self.file.write(header_json)
            self.file.write(b' ' * overhead)
            self.file.close()


DTYPE_MAPPING = {
    'F64': (torch.float64, 8),
    'F32': (torch.float32, 4),
    'F16': (torch.float16, 2),
    'BF16': (torch.bfloat16, 2),
    'I8': (torch.int8, 1),
    'I64': (torch.int64, 8),
    'I32': (torch.int32, 4),
    'I16': (torch.int16, 2),
}


DTYPE_REVERSE_MAPPING = {v: (k, b) for k, (v, b) in DTYPE_MAPPING.items()}



import dataclasses
import functools
import gc
import logging
import pathlib
import threading
import torch
from concurrent.futures import ThreadPoolExecutor, as_completed
from sd_mecha.model_detection import DetermineConfigVisitor
from sd_mecha.recipe_nodes import RecipeVisitor
from sd_mecha.streaming import InSafetensorsDict, OutSafetensorsDict
from sd_mecha import extensions, recipe_nodes, recipe_serializer
from tqdm import tqdm
from typing import Optional, Mapping, MutableMapping, Dict, Set, List


class RecipeMerger:
    def __init__(
        self, *,
        models_dir: Optional[pathlib.Path | str | List[pathlib.Path | str]] = None,
        default_device: str = "cpu",
        default_dtype: Optional[torch.dtype] = torch.float64,
        tqdm: type = tqdm,
    ):
        if models_dir is None:
            models_dir = []
        if not isinstance(models_dir, List):
            models_dir = [models_dir]
        for i in range(len(models_dir)):
            if isinstance(models_dir[i], str):
                models_dir[i] = pathlib.Path(models_dir[i])
            if models_dir[i] is not None:
                models_dir[i] = models_dir[i].absolute()

        self.__base_dirs = models_dir

        self.__default_device = default_device
        self.__default_dtype = default_dtype
        self.__tqdm = tqdm

    def merge_and_save(
        self, recipe: extensions.merge_method.RecipeNodeOrPath, *,
        output: MutableMapping[str, torch.Tensor] | pathlib.Path | str = "merge",
        fallback_model: Optional[Mapping[str, torch.Tensor] | recipe_nodes.ModelRecipeNode | pathlib.Path | str] = None,
        save_device: Optional[str] = "cpu",
        save_dtype: Optional[torch.dtype] = torch.float16,
        threads: Optional[int] = None,
        total_buffer_size: int = 2**28,
    ):
        recipe = extensions.merge_method.path_to_node(recipe)
        if recipe.merge_space != recipe_nodes.MergeSpace.BASE:
            raise ValueError(f"recipe should be in model merge space, not {str(recipe.merge_space).split('.')[-1]}")
        if isinstance(fallback_model, (str, pathlib.Path)):
            fallback_model = extensions.merge_method.path_to_node(fallback_model)
        elif not isinstance(fallback_model, (recipe_nodes.ModelRecipeNode, Mapping, type(None))):
            raise ValueError(f"fallback_model should be a simple model or None, not {type(fallback_model)}")
        extensions.merge_method.clear_model_paths_cache()

        fallback_is_recipe = isinstance(fallback_model, recipe_nodes.ModelRecipeNode)
        fallback_in_recipe = fallback_is_recipe and fallback_model in recipe
        total_files_open = (
            recipe.accept(recipe_nodes.ModelsCountVisitor()) +
            int(isinstance(output, (str, pathlib.Path))) +
            int(fallback_is_recipe and not fallback_in_recipe)
        )
        buffer_size_per_file = total_buffer_size // total_files_open
        if threads is None:
            threads = total_files_open

        load_input_dicts_visitor = LoadInputDictsVisitor(
            self.__base_dirs,
            buffer_size_per_file,
        )
        recipe.accept(load_input_dicts_visitor)
        if fallback_is_recipe:
            fallback_model.accept(load_input_dicts_visitor)

        model_config = recipe.accept(DetermineConfigVisitor())
        if fallback_is_recipe:
            model_config = model_config.intersect(fallback_model.accept(DetermineConfigVisitor()))
            fallback_model = fallback_model.state_dict

        output = self.__normalize_output_to_dict(
            output,
            model_config.get_minimal_dummy_header(),
            model_config.get_keys_to_merge(),
            recipe_serializer.serialize(recipe),
            buffer_size_per_file // threads,
            save_dtype,
        )

        thread_local_data = threading.local()
        progress = self.__tqdm(total=len(model_config.keys()), desc="Merging recipe")
        with ThreadPoolExecutor(max_workers=threads) as executor:
            futures = []
            for key in model_config.keys():
                key_merger = model_config.get_key_merger(key, recipe, fallback_model, self.__default_device, self.__default_dtype)
                key_merger = self.__track_output(key_merger, output, key, save_dtype, save_device)
                key_merger = self.__track_progress(key_merger, key, model_config.get_shape(key), progress)
                key_merger = self.__wrap_thread_context(key_merger, thread_local_data)
                futures.append(executor.submit(key_merger))

            for future in as_completed(futures):
                if future.exception() is not None:
                    for future_to_cancel in futures:
                        future_to_cancel.cancel()
                    raise future.exception()
                future.result()

        progress.close()
        if isinstance(output, OutSafetensorsDict):
            output.close()
        recipe.accept(CloseInputDictsVisitor())

        gc.collect()
        torch.cuda.empty_cache()

    def __normalize_output_to_dict(
        self,
        output: MutableMapping[str, torch.Tensor] | pathlib.Path | str,
        merged_header: Dict[str, dict],
        keys_to_merge: Set[str],
        serialized_recipe: str,
        buffer_size_per_thread: int,
        dtype: torch.dtype,
    ):
        if isinstance(output, (str, pathlib.Path)):
            if not isinstance(output, pathlib.Path):
                output = pathlib.Path(output)
            if not output.is_absolute():
                output = self.__base_dirs[0] / output
            if not output.suffix:
                output = output.with_suffix(".safetensors")
            logging.info(f"Saving to {output}")

            output = OutSafetensorsDict(
                output,
                merged_header,
                keys_to_merge,
                serialized_recipe,
                buffer_size_per_thread,
                dtype,
            )
        return output

    def __track_progress(self, f, key, key_shape, progress):
        @functools.wraps(f)
        def track_progress(*args, **kwargs):
            progress.set_postfix({"key": key, "shape": key_shape})
            res = f(*args, **kwargs)
            progress.update()
            return res
        return track_progress

    def __track_output(self, f, output, key, save_dtype, save_device):
        if save_dtype is None:
            save_dtype = self.__default_dtype

        if save_device is None:
            to_kwargs = {"dtype": save_dtype},
        else:
            to_kwargs = {"dtype": save_dtype, "device": save_device}

        @functools.wraps(f)
        def track_output(*args, **kwargs):
            output[key] = f(*args, **kwargs).to(**to_kwargs)
        return track_output

    def __wrap_thread_context(self, f, ctx):
        @functools.wraps(f)
        def thread_context(*args, **kwargs):
            if torch.cuda.is_available():
                if not hasattr(ctx, 'cuda_stream'):
                    ctx.cuda_stream = torch.cuda.Stream()
                with torch.cuda.stream(ctx.cuda_stream):
                    return f(*args, **kwargs)
            else:
                return f(*args, **kwargs)

        return thread_context


@dataclasses.dataclass
class LoadInputDictsVisitor(RecipeVisitor):
    __base_dirs: List[pathlib.Path]
    __buffer_size_per_dict: int

    def visit_model(self, node: recipe_nodes.ModelRecipeNode):
        node.state_dict = self.__load_dict(node)

    def visit_parameter(self, _node: recipe_nodes.ParameterRecipeNode):
        return

    def visit_merge(self, node: recipe_nodes.MergeRecipeNode):
        for model in node.models:
            model.accept(self)

    def __load_dict(
        self,
        node: recipe_nodes.ModelRecipeNode,
    ) -> InSafetensorsDict:
        if node.state_dict is not None:
            return node.state_dict

        path = node.path
        if not isinstance(path, pathlib.Path):
            path = pathlib.Path(path)
        if not path.is_absolute():
            for base_dir in self.__base_dirs:
                path_attempt = base_dir / path
                if not path_attempt.suffix:
                    path_attempt = path_attempt.with_suffix(".safetensors")
                if path_attempt.exists():
                    path = path_attempt
                    break

        return InSafetensorsDict(path, self.__buffer_size_per_dict)


@dataclasses.dataclass
class CloseInputDictsVisitor:
    def visit_model(self, node: recipe_nodes.ModelRecipeNode):
        if node.state_dict is not None:
            node.state_dict.close()
        node.state_dict = None

    def visit_parameter(self, _node: recipe_nodes.ParameterRecipeNode):
        return

    def visit_merge(self, node: recipe_nodes.MergeRecipeNode):
        for model in node.models:
            model.accept(self)



import abc
import pathlib
import torch
from typing import Optional, Dict, Any, Mapping

from torch import Tensor

from sd_mecha import extensions
from sd_mecha.extensions.model_arch import ModelArch
from sd_mecha.hypers import validate_hyper, Hyper
from sd_mecha.merge_space import MergeSpace


class RecipeNode(abc.ABC):
    @abc.abstractmethod
    def accept(self, visitor, *args, **kwargs):
        pass

    @property
    @abc.abstractmethod
    def merge_space(self) -> MergeSpace:
        pass

    @property
    @abc.abstractmethod
    def model_arch(self) -> Optional[ModelArch]:
        pass

    @abc.abstractmethod
    def __contains__(self, item):
        pass


class ModelRecipeNode(RecipeNode):
    def __init__(
        self,
        state_dict: str | pathlib.Path | Mapping[str, Tensor],
        model_arch: str = "sd1",
        model_type: str = "base",
    ):
        if isinstance(state_dict, Mapping):
            self.path = None
            self.state_dict = state_dict
        else:
            self.path = state_dict
            self.state_dict = None
        self.model_type = extensions.model_type.resolve(model_type, model_arch)
        self.__model_arch = extensions.model_arch.resolve(model_arch)

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_model(self, *args, **kwargs)

    @property
    def merge_space(self) -> MergeSpace:
        return self.model_type.merge_space

    @property
    def model_arch(self) -> Optional[ModelArch]:
        return self.__model_arch

    def __contains__(self, item):
        if isinstance(item, ModelRecipeNode):
            return self.path == item.path
        else:
            return False


class ParameterRecipeNode(RecipeNode):
    def __init__(self, name: str, merge_space: MergeSpace, model_arch: Optional[str] = None):
        self.name = name
        self.__merge_space = merge_space
        if model_arch is not None:
            model_arch = extensions.model_arch.resolve(model_arch)
        self.__model_arch = model_arch

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_parameter(self, *args, **kwargs)

    @property
    def merge_space(self) -> MergeSpace:
        return self.__merge_space

    @property
    def model_arch(self) -> Optional[ModelArch]:
        return self.__model_arch

    def __contains__(self, item):
        if isinstance(item, ParameterRecipeNode):
            return self.name == item.name and self.merge_space == item.merge_space
        else:
            return False


class MergeRecipeNode(RecipeNode):
    def __init__(
        self,
        merge_method,
        *models: RecipeNode,
        hypers: Dict[str, Hyper],
        volatile_hypers: Dict[str, Any],
        device: Optional[str] = None,
        dtype: Optional[torch.dtype] = None,
    ):
        self.merge_method = merge_method
        self.models = models
        for hyper_v in hypers.values():
            validate_hyper(hyper_v, self.model_arch)
        self.hypers = hypers
        self.volatile_hypers = volatile_hypers
        self.device = device
        self.dtype = dtype
        self.__merge_space = self.merge_method.get_return_merge_space([
            model.merge_space
            for model in self.models
        ])

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_merge(self, *args, **kwargs)

    @property
    def merge_space(self) -> MergeSpace:
        return self.__merge_space

    @property
    def model_arch(self) -> Optional[ModelArch]:
        if not self.models:
            return None

        for model in self.models:
            if (arch := model.model_arch) is None:
                return None

        return arch

    def __contains__(self, item):
        if isinstance(item, MergeRecipeNode):
            return self is item or any(
                item in model
                for model in self.models
            )
        else:
            return False


class RecipeVisitor(abc.ABC):
    @abc.abstractmethod
    def visit_model(self, node: ModelRecipeNode):
        pass

    @abc.abstractmethod
    def visit_parameter(self, node: ParameterRecipeNode):
        pass

    @abc.abstractmethod
    def visit_merge(self, node: MergeRecipeNode):
        pass


class DepthRecipeVisitor(RecipeVisitor):
    def visit_model(self, _node: ModelRecipeNode):
        return 1

    def visit_parameter(self, _node: ParameterRecipeNode):
        return 0

    def visit_merge(self, node: MergeRecipeNode):
        return max(
            model.accept(self)
            for model in node.models
        ) + 1


class ModelsCountVisitor(RecipeVisitor):
    def __init__(self):
        self.__seen_nodes = []

    def visit_model(self, node: ModelRecipeNode):
        seen = node in self.__seen_nodes
        self.__seen_nodes.append(node)
        return not seen

    def visit_parameter(self, _node: ParameterRecipeNode):
        return 0

    def visit_merge(self, node: MergeRecipeNode):
        return sum(
            model.accept(self)
            for model in node.models
        )


class ParameterResolverVisitor(RecipeVisitor):
    def __init__(self, arguments: Dict[str, RecipeNode]):
        self.__arguments = arguments

    def visit_model(self, node: ModelRecipeNode) -> RecipeNode:
        return node

    def visit_parameter(self, node: ParameterRecipeNode) -> RecipeNode:
        return self.__arguments.get(node.name, node)

    def visit_merge(self, node: MergeRecipeNode) -> RecipeNode:
        return MergeRecipeNode(
            node.merge_method,
            *(node.accept(self) for node in node.models),
            hypers=node.hypers,
            volatile_hypers=node.volatile_hypers,
            device=node.device,
            dtype=node.dtype,
        )



import json
import pathlib
import torch
from sd_mecha.merge_space import MergeSpace
from sd_mecha.extensions.model_type import register_model_type
from typing import Mapping


@register_model_type(merge_space=MergeSpace.BASE, needs_header_conversion=False)
def base(state_dict: Mapping[str, torch.Tensor], key: str, **kwargs) -> torch.Tensor:
    return state_dict[key]


@register_model_type(merge_space=MergeSpace.DELTA, key_suffixes=[".lora_up.weight", ".lora_down.weight", ".alpha"], strict_suffixes=True)
def lora(state_dict: Mapping[str, torch.Tensor], key: str, **kwargs) -> torch.Tensor:
    if key.endswith(".bias"):
        raise KeyError(key)

    if key.startswith("cond_stage_model.transformer."):
        lora_key = "lora_te_" + "_".join(key.split(".")[2:-1])
    else:
        lora_key = SD1_MODEL_TO_LORA_KEYS[key]

    return compose_lora_up_down(state_dict, lora_key)


with open(pathlib.Path(__file__).parent / "lora" / "sd1_ldm_to_lora.json", 'r') as f:
    SD1_MODEL_TO_LORA_KEYS = json.load(f)


@register_model_type(merge_space=MergeSpace.DELTA, model_archs="sdxl", key_suffixes=[".lora_up.weight", ".lora_down.weight", ".alpha"], strict_suffixes=True)
def lora(state_dict: Mapping[str, torch.Tensor], key: str, **kwargs) -> torch.Tensor:
    if key.endswith((".bias", "_bias")):
        raise KeyError(key)

    if key.startswith("model.diffusion_model."):
        lora_key = "lora_unet_" + "_".join(key.split(".")[2:-1])
        return compose_lora_up_down(state_dict, lora_key)
    elif key.startswith("conditioner.embedders.0.transformer."):
        lora_key = "lora_te1_" + "_".join(key.split(".")[4:-1])
        return compose_lora_up_down(state_dict, lora_key)
    elif key.startswith("conditioner.embedders.1.model.transformer.resblocks."):
        # [6:] instead of [6:-1] because `key` can end with either ".weight" or "_weight"
        lora_key = "lora_te2_text_model_encoder_layers_" + "_".join(key.split(".")[6:])
        lora_key = lora_key.replace("_attn_", "_self_attn_")
        lora_key = lora_key.replace("_mlp_c_fc_", "_mlp_fc1_")
        lora_key = lora_key.replace("_mlp_c_proj_", "_mlp_fc2_")
        lora_key = lora_key.replace("_weight", "")

        if lora_key.endswith("_in_proj"):
            lora_key = lora_key.replace("_in_proj", "")
            lora_keys = [
                f"{lora_key}_{k}_proj"
                for k in ("q", "k", "v")
            ]
            return torch.vstack([
                compose_lora_up_down(state_dict, k)
                for k in lora_keys
            ])

        return compose_lora_up_down(state_dict, lora_key)
    else:
        raise KeyError(key)


def compose_lora_up_down(state_dict: Mapping[str, torch.Tensor], key: str):
    up_weight = state_dict[f"{key}.lora_up.weight"].to(torch.float64)
    down_weight = state_dict[f"{key}.lora_down.weight"].to(torch.float64)
    alpha = state_dict[f"{key}.alpha"].to(torch.float64)
    dim = down_weight.size()[0]

    if len(down_weight.size()) == 2:  # linear
        res = up_weight @ down_weight
    elif down_weight.size()[2:4] == (1, 1):  # conv2d 1x1
        res = (up_weight.squeeze(3).squeeze(2) @ down_weight.squeeze(3).squeeze(2)).unsqueeze(2).unsqueeze(3)
    else:  # conv2d 3x3
        res = torch.nn.functional.conv2d(down_weight.permute(1, 0, 2, 3), up_weight).permute(1, 0, 2, 3)
    return res * (alpha / dim)



import abc
import dataclasses
import functools
import pathlib
import torch
from typing import Iterable, Dict, List, Mapping, Callable, Tuple, Optional
from sd_mecha.extensions.model_arch import ModelArch
from sd_mecha.hypers import get_hyper
from sd_mecha.recipe_nodes import RecipeNode, ModelRecipeNode, ParameterRecipeNode, MergeRecipeNode, DepthRecipeVisitor, RecipeVisitor
from sd_mecha.streaming import InSafetensorsDict


@dataclasses.dataclass
class ModelConfig:
    __minimal_dummy_header: Dict[str, Dict[str, str | List[int]]]
    __input_paths: List[str | pathlib.Path]
    __model_arch: ModelArch

    def __post_init__(self):
        self.__keys_to_forward = set(
            k for k in self.__minimal_dummy_header
            if k in self.__model_arch.keys_to_forward
        )
        self.__keys_to_merge = set(
            k for k in self.__minimal_dummy_header
            if k not in self.__keys_to_forward
            and k in self.__model_arch.keys_to_merge
        )
        self.__minimal_dummy_header = {
            k: v
            for k, v in self.__minimal_dummy_header.items()
            if k in self.__keys_to_forward or k in self.__keys_to_merge
        }

    def keys(self) -> Iterable[str]:
        return self.get_minimal_dummy_header().keys()

    def get_keys_to_merge(self):
        return self.__keys_to_merge

    def get_shape(self, key: str) -> List[int]:
        return self.get_minimal_dummy_header()[key]["shape"]

    def get_minimal_dummy_header(self) -> Dict[str, Dict[str, str | List[int]]]:
        return self.__minimal_dummy_header

    def get_input_paths(self) -> List[str | pathlib.Path]:
        return self.__input_paths

    def intersect(self, other):
        if self.__model_arch is not other.__model_arch:
            self_paths = ', '.join(str(p) for p in self.__input_paths)
            other_paths = ', '.join(str(p) for p in other.__input_paths)
            raise ValueError(
                "Found incompatible model architectures: "
                f"{len(self.__input_paths)} {self.__model_arch} models ({self_paths}) and "
                f"{len(other.__input_paths)} {other.__model_arch} models ({other_paths})"
            )

        return ModelConfig(
            self.__minimal_dummy_header | other.__minimal_dummy_header,
            self.__input_paths + other.__input_paths,
            self.__model_arch,
        )

    def get_key_merger(
        self,
        key: str,
        recipe: RecipeNode,
        fallback_model: Mapping[str, torch.Tensor],
        device: str,
        dtype: torch.dtype,
    ) -> Callable[[], torch.Tensor]:
        key_merger = KeyMerger(
            recipe,
            fallback_model,
            device,
            dtype,
        )
        if key in self.__keys_to_forward:
            return functools.partial(key_merger.forward_and_save, key)
        else:
            return functools.partial(key_merger.merge_and_save, key)


class DetermineConfigVisitor(RecipeVisitor):
    def visit_model(self, node: ModelRecipeNode) -> ModelConfig:
        state_dict_path = getattr(node.state_dict, "file_path", "<memory>")
        return ModelConfig(
            node.model_type.convert_header(state_dict_path, node.state_dict, node.model_arch),
            [state_dict_path],
            node.model_arch,
        )

    def visit_parameter(self, node: ParameterRecipeNode) -> ModelConfig:
        raise TypeError("Recipe parameters do not have configuration")

    def visit_merge(self, node: MergeRecipeNode) -> ModelConfig:
        configs = [
            model.accept(self)
            for model in node.models
        ]
        if configs:
            return functools.reduce(ModelConfig.intersect, configs)
        raise ValueError("No input models")


@dataclasses.dataclass
class KeyMerger:
    recipe: RecipeNode
    fallback_model: Mapping[str, torch.Tensor]
    default_device: str
    default_dtype: torch.dtype

    def merge_and_save(
        self,
        key: str,
    ) -> torch.Tensor:
        key_merger = KeyMergeVisitor(
            key,
            self.default_device,
            self.default_dtype,
            self.__get_passthrough_tensor,
        )
        return self.recipe.accept(key_merger)

    def forward_and_save(
        self,
        key: str,
    ) -> torch.Tensor:
        return self.__get_passthrough_tensor(key)

    def __get_passthrough_tensor(self, key: str):
        if self.fallback_model is not None and key in self.fallback_model:
            return self.fallback_model[key]

        key_merger = KeyPassthroughVisitor(
            key,
            self.default_device,
            self.default_dtype,
        )
        return self.recipe.accept(key_merger)


@dataclasses.dataclass
class KeyVisitor(RecipeVisitor, abc.ABC):
    _key: str
    _default_device: str
    _default_dtype: torch.dtype

    def visit_model(self, node: ModelRecipeNode) -> torch.Tensor:
        state_dict_path = getattr(node.state_dict, "file_path", "<memory>")
        return node.model_type.get_tensor(state_dict_path, node.state_dict, self._key)

    def visit_parameter(self, node: ParameterRecipeNode) -> torch.Tensor:
        raise NotImplementedError(f"Interactive arguments are not yet implemented. Recipe is abstract: parameter '{node.name}' has no value.")

    @abc.abstractmethod
    def visit_merge(self, node: MergeRecipeNode) -> torch.Tensor:
        pass


@dataclasses.dataclass
class KeyMergeVisitor(KeyVisitor):
    _passthrough_callback: Callable[[str], torch.Tensor]

    def visit_merge(self, node: MergeRecipeNode) -> torch.Tensor:
        merged: List[Optional[torch.Tensor]] = [None] * len(node.models)
        try:
            return node.merge_method(
                self.__visit_deeper_first(node.models, merged),
                {k: get_hyper(v, self._key, node.model_arch) for k, v in node.hypers.items()} | node.volatile_hypers,
                self._key,
                node.device if node.device is not None else self._default_device,
                node.dtype if node.dtype is not None else self._default_dtype,
            )
        except KeyError:
            for n, m in ((n, m) for n, m in zip(node.models, merged) if m is not None):
                if n.merge_space == node.merge_space:
                    return m
            return self._passthrough_callback(self._key)

    def __visit_deeper_first(self, nodes: Tuple[RecipeNode, ...], merged: List[Optional[torch.Tensor]]) -> list:
        def depth_of_value(index) -> int:
            if nodes[index] is None:
                return 0
            return nodes[index].accept(DepthRecipeVisitor())

        for index in sorted(range(len(nodes)), key=depth_of_value, reverse=True):
            if nodes[index] is None:
                continue
            merged[index] = nodes[index].accept(self)

        return merged


@dataclasses.dataclass
class KeyPassthroughVisitor(KeyVisitor):
    def visit_merge(self, node: MergeRecipeNode) -> torch.Tensor:
        for model in node.models:
            try:
                return model.accept(self)
            except KeyError:
                continue

        raise KeyError(f"No model has key '{self._key}'")
